import itertools as it
from typing import Optional

import matplotlib.pyplot as plt
import networkx as nx

from mrtommy.scene import Point2D, Scene, Segment2D


class Solver:
    """
    A solver should be able to solve a specific Scene.
    Any subclass must override the solve function, and may set the roadmap if relevant to some graph where each vertex is a tuple of Point2D.

    The solver can also be checked and plotted using the check and plot functions accordingly.
    """

    roadmap: Optional[nx.Graph]

    def solve(self, scene: Scene) -> Optional[list[list[Point2D]]]:
        """
        Solves the motion planning problem by giving a list of points to travel through for each robot,
        where the robots travel in straight lines (each element is a list of Point2D for every robot)
        """
        pass

    def check(self, scene: Scene, solution: Optional[list[list[Point2D]]]):
        """Checks the given solution is valid and there are no intersections"""

        if solution is None:
            return False

        for s in solution:
            if len(scene.robots) != len(s):
                raise Exception(
                    "Each list of points in the solution must be a location for each robot!"
                )

        for robot, start in zip(scene.robots, solution[0]):
            if start != robot.start:
                return False

        for robot, end in zip(scene.robots, solution[-1]):
            if end != robot.end:
                return False

        for i in range(len(solution) - 1):
            for robot_index in range(len(scene.robots)):
                robot = scene.robots[robot_index]
                path = Segment2D(
                    solution[i][robot_index], solution[i + 1][robot_index]
                )
                for obstacle in scene.obstacles:
                    if (
                        path.distance(obstacle.location)
                        < robot.radius + obstacle.radius
                    ):
                        return False

            for r1, r2 in it.combinations(range(len(scene.robots)), 2):
                p1 = Segment2D(solution[i][r1], solution[i + 1][r2])
                p2 = Segment2D(solution[i][r1], solution[i + 1][r2])
                robot1 = scene.robots[r1]
                robot2 = scene.robots[r2]
                if p1.distance(p2) < robot1.radius + robot2.radius:
                    return False

        return True

    def solve_and_check(self, scene: Scene):
        """Solves and checks the solution of the given scene."""

        solution = self.solve(scene)
        if self.check(scene, solution):
            print("Solution is valid!")
        else:
            print("Solution invalid!")

        return solution

    def plot(
        self,
        scene: Scene,
        paths: Optional[list[list[Point2D]]] = None,
        filename: Optional[str] = None,
        display=False,
        plot_roadmap=True,
        title=True,
        robot_colors: list[str] = [
            "yellow",
            "green",
            "cyan",
            "blue",
            "indigo",
            "fuchsia",
        ],
        obstacle_color="red",
    ):
        """
        Plots the scene, and the given solution `paths` if available.

        If `display` is set, opens a matplotlib window with the plot. If `filename` is set, saves the plot to the given filename.
        """

        plt.clf()
        plt.plot(
            [scene.x_min, scene.x_max, scene.x_max, scene.x_min, scene.x_min],
            [scene.y_min, scene.y_min, scene.y_max, scene.y_max, scene.y_min],
            color="black",
        )

        for obstacle in scene.obstacles:
            circle = plt.Circle(
                (obstacle.location.x, obstacle.location.y),
                obstacle.radius,
                color=obstacle_color,
            )
            plt.gca().add_patch(circle)

        if plot_roadmap and self.roadmap is not None:
            for s, e in self.roadmap.edges:
                for p1, p2 in zip(s, e):
                    plt.plot([p1.x, p2.x], [p1.y, p2.y], color="black")

        if paths is not None:
            for robot_index in range(len(scene.robots)):
                for i in range(len(paths) - 1):
                    plt.plot(
                        [paths[i][robot_index].x, paths[i + 1][robot_index].x],
                        [paths[i][robot_index].y, paths[i + 1][robot_index].y],
                        linewidth=3,
                        color=robot_colors[robot_index],
                    )

        for index, robot in enumerate(scene.robots):
            circle = plt.Circle(
                (robot.start.x, robot.start.y),
                robot.radius,
                color=robot_colors[index],
            )
            plt.gca().add_patch(circle)
            circle = plt.Circle(
                (robot.end.x, robot.end.y),
                robot.radius,
                color=robot_colors[index],
            )
            plt.gca().add_patch(circle)

        if title:
            plt.title(str(self))

        if filename is not None:
            plt.savefig(filename)
        if display:
            plt.show()



import math
from typing import Optional, Union

from mrtommy.constants import EPSILON


class Point2D:
    x: float
    y: float

    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y

    def distance(self, other: "Point2D"):
        return (self - other).norm()

    def __mul__(self, other: "Point2D"):
        return self.x * other.x + self.y * other.y

    def __sub__(self, other: "Point2D"):
        return Point2D(self.x - other.x, self.y - other.y)

    def norm(self):
        return math.sqrt(self.x**2 + self.y**2)

    def normalized(self):
        norm = self.norm()
        return Point2D(self.x / norm, self.y / norm)


class Line2D:
    start: Point2D
    end: Point2D

    def __init__(self, start: Point2D, end: Point2D):
        self.start = start
        self.end = end

    def representation(self) -> tuple[float, float, float]:
        """Returns the a,b,c in the representation of the line as ax+by+c=0"""

        if self.start.x == self.end.x:
            return 1, 0, self.start.x

        slope = (self.end.y - self.start.y) / (self.end.x - self.start.x)
        return slope, -1, (self.start.y - slope * self.start.x)

    def distance(self, point: Point2D):
        """Returns the distance between this line and the given point"""

        a, b, c = self.representation()
        return abs(a * point.x + b * point.y + c) / math.sqrt(a * a + b * b)

    def intersection(self, other: "Line2D") -> Optional[Point2D]:
        """Returns the intersection of these two lines, or None if there isn't any"""

        a1, b1, c1 = self.representation()
        a2, b2, c2 = other.representation()

        denominator = b1 * a2 - a1 * b2
        if denominator == 0 or a1 == 0 or a2 == 0:
            return None

        numerator = a1 * c2 - c1 * a2
        y = numerator / denominator
        x = (-b1 * y - c1) / a1
        return Point2D(x, y)

    def contains(self, point: Point2D):
        """Checks whether the given point is on this line"""

        a, b, c = self.representation()
        return abs(a * point.x + b * point.y + c) < EPSILON

    def perpendicular(self, point: Point2D):
        """Returns the line perpendicular to this line passing at the given point"""

        a, b, c = self.representation()
        offset = a * point.y - b * point.x
        return line_from_representation(b, -a, offset)

    def __eq__(self, other: "Line2D"):
        return self.representation() == other.representation()


def line_from_representation(a: float, b: float, c: float) -> Line2D:
    """Initializes the line from the given ax+by+c=0 representation"""

    if a == 0:
        return Line2D(Point2D(0, -c / b), Point2D(1, -c / b))

    if b == 0:
        return Line2D(Point2D(-c / a, 0), Point2D(-c / a, 1))

    return Line2D(Point2D(1, -(a + c) / b), Point2D(0, -c / b))


class Segment2D:
    start: Point2D
    end: Point2D

    def __init__(self, start: Point2D, end: Point2D):
        self.start = start
        self.end = end

    def line(self):
        return Line2D(self.start, self.end)

    def length(self):
        return (self.end - self.start).norm()

    def contains(self, point: Point2D):
        """Checks whether the given point is inside this segment"""

        if not self.line().contains(point):
            return False

        dot_product = (point - self.start) * (self.end - self.start)
        if dot_product < 0 or dot_product > self.length() ** 2:
            return False

        return True

    def intersection(
        self, other: Union[Line2D, "Segment2D"]
    ) -> Optional[Point2D]:
        if isinstance(other, Segment2D):
            p = self.line().intersection(other.line())
            if p is not None and self.contains(p) and other.contains(p):
                return p
            return None

        assert isinstance(other, Line2D)
        p = self.line().intersection(other)
        if p is not None and self.contains(p):
            return p
        return None

    def distance(self, other: Union[Point2D, Line2D, "Segment2D"]):
        """Returns the distance between this segment and the given object"""

        if isinstance(other, Point2D):
            closest = self.intersection(self.line().perpendicular(other))
            if closest is not None:
                return closest.distance(other)

            return min(
                self.start.distance(other),
                self.end.distance(other),
            )

        if isinstance(other, Segment2D):
            if self.intersection(other) is not None:
                return 0

            return min(
                self.distance(other.start),
                self.distance(other.end),
                other.distance(self.start),
                other.distance(self.end),
            )

        assert isinstance(other, Line2D)
        if self.intersection(other) is not None:
            return 0
        return min(other.distance(self.start), other.distance(self.end))



import networkx as nx

from mrtommy.scene import Point2D, Scene
from mrtommy.solver import Solver
from mrtommy.utilities import distance


class RRTSolver(Solver):
    def __init__(self, max_iterations: int, max_distance: float):
        self.max_iterations = max_iterations
        self.max_distance = max_distance

    def move(self, start: Point2D, end: Point2D, max_distance: float):
        if start.distance(end) <= max_distance:
            return end

        difference = (end - start).normalized()

        return Point2D(
            start.x + difference.x * max_distance,
            start.y + difference.y * max_distance,
        )

    def get_position(self, scene: Scene):
        return scene.get_random_positions()

    def solve(self, scene):
        self.roadmap = nx.Graph()
        start = tuple(scene.start)
        end = tuple(scene.end)

        self.roadmap.add_node(start)

        for _ in range(self.max_iterations):
            positions = self.get_position(scene)

            closest = min(
                self.roadmap.nodes,
                key=lambda n: distance(positions, n),
            )
            positions = [
                self.move(c, p, self.max_distance)
                for c, p in zip(closest, positions)
            ]

            if not scene.check_edge(closest, positions):
                continue

            # make positions hashable
            positions = tuple(positions)

            self.roadmap.add_node(positions)
            self.roadmap.add_edge(
                closest, positions, weight=distance(closest, positions)
            )

        self.roadmap.add_node(end)
        for positions in self.roadmap.nodes:
            if scene.check_edge(positions, end):
                self.roadmap.add_edge(
                    positions, end, weight=distance(positions, end)
                )

        if not nx.algorithms.has_path(self.roadmap, start, end):
            return None

        return nx.algorithms.shortest_path(
            self.roadmap, start, end, weight="weight"
        )

    def __repr__(self):
        return f"RRT with iterations={self.max_iterations} and η={self.max_distance}"



import networkx as nx

from mrtommy.solver import Solver
from mrtommy.utilities import distance, nearest_neighbors


class PRMSolver(Solver):
    def __init__(self, landmarks: int, k: int):
        self.landmarks = landmarks
        self.k = k

    def solve(self, scene):
        self.roadmap = nx.Graph()

        start = tuple(scene.start)
        end = tuple(scene.end)

        self.roadmap.add_node(start)
        self.roadmap.add_node(end)

        for _ in range(self.landmarks):
            self.roadmap.add_node(tuple(scene.get_random_valid_positions()))

        for v in self.roadmap.nodes:
            for neighbor in nearest_neighbors(v, self.roadmap.nodes, self.k):
                if scene.check_edge(v, neighbor):
                    self.roadmap.add_edge(
                        v, neighbor, weight=distance(v, neighbor)
                    )

        if not nx.algorithms.has_path(self.roadmap, start, end):
            return None

        return nx.algorithms.shortest_path(
            self.roadmap, start, end, weight="weight"
        )

    def __repr__(self):
        return f"PRM with landmarks={self.landmarks} and k={self.k}"



import itertools as it

import networkx as nx

from mrtommy.scene import RobotDisc, Scene
from mrtommy.solver import Solver
from mrtommy.utilities import distance


class HGraphSolver(Solver):
    def __init__(
        self,
        solvers: list[Solver],
        local_solver: Solver,
        neighborhood_distance: float,
    ):
        self.solvers = solvers
        self.local_solver = local_solver
        self.neighborhood_distance = neighborhood_distance

    def solve(self, scene):
        self.roadmap = nx.Graph()
        start = tuple(scene.start)
        end = tuple(scene.end)

        paths = []
        for solver in self.solvers:
            path = solver.solve(scene)
            if path is not None:
                paths.append(path)
                for i in range(len(path) - 1):
                    self.roadmap.add_edge(path[i], path[i + 1])

        # TODO: implement All-Pairs H-Graph or Edit-Distance H-Graph
        for p1, p2 in it.combinations(paths, 2):
            for v1 in p1:
                for v2 in p2:
                    # Neighborhood
                    if distance(v1, v2) > self.neighborhood_distance:
                        continue

                    local_scene = Scene(
                        [
                            RobotDisc(start, end, robot.radius)
                            for robot, start, end in zip(scene.robots, v1, v2)
                        ],
                        scene.obstacles,
                    )
                    local_path = self.local_solver.solve(local_scene)
                    if local_path is not None:
                        for i in range(len(local_path) - 1):
                            self.roadmap.add_edge(
                                local_path[i], local_path[i + 1]
                            )

        if not nx.algorithms.has_path(self.roadmap, start, end):
            return None

        return nx.algorithms.shortest_path(
            self.roadmap, start, end, weight="weight"
        )

    def __repr__(self):
        return (
            f"HGraph with local={self.local_solver.__class__.__name__} and "
            + ", ".join([s.__class__.__name__ for s in self.solvers])
        )



from mrtommy.scene import load_scene

if __name__ == "__main__":
    scenes = [
        load_scene("../resources/coffee_shop.json"),
        load_scene("../resources/2_coffee_shop.json"),
        load_scene("../resources/big_coffee_shop.json"),
    ]

    # run_solver(
    #     PRMSolver,
    #     {
    #         "landmarks": range(5000, 10100, 100),
    #         "k": [5, 10, 15],
    #     },
    #     scenes[2],
    # )

    # run_solver(
    #     RRTSolver,
    #     {
    #         "max_iterations": range(3000, 7100, 100),
    #         "max_distance": [1.5],
    #     },
    #     scenes[2],
    # )

    # run_solver(
    #     RRTStarSolver,
    #     {
    #         "max_iterations": range(3000, 7100, 100),
    #         "max_distance": [1.5],
    #         "radius": [3],
    #     },
    #     scenes[2],
    # )

    # run_solver(
    #     DRRTSolver,
    #     {
    #         "prm_landmarks": [2000, 4000],
    #         "prm_k": [0],
    #         "max_iterations": range(3000, 7100, 100),
    #         "max_distance": [1.5],
    #     },
    #     scenes[2],
    # )

    # run_solver(
    #     DRRTStarSolver,
    #     {
    #         "prm_landmarks": [1000],
    #         "prm_k": [0],
    #         "max_iterations": range(3000, 7100, 100),
    #         "max_distance": [1.5],
    #         "radius": [3],
    #     },
    #     scenes[2],
    # )

    # run_solver(
    #     HGraphSolver,
    #     {
    #         "solvers": [
    #             [PRMSolver(3000, 5) for i in range(2)],
    #             [PRMSolver(3000, 5) for i in range(3)],
    #             [PRMSolver(3000, 5) for i in range(5)],
    #             [PRMSolver(3000, 5) for i in range(10)],
    #             [RRTSolver(3000, 1.5) for i in range(2)],
    #             [RRTSolver(3000, 1.5) for i in range(3)],
    #             [RRTSolver(3000, 1.5) for i in range(5)],
    #             [RRTSolver(3000, 1.5) for i in range(10)],
    #             [RRTStarSolver(3000, 0.5, i) for i in range(2)],
    #             [RRTStarSolver(3000, 0.5, i) for i in range(3)],
    #             [RRTStarSolver(3000, 0.5, i) for i in range(5)],
    #             [PRMSolver(3000, 5) for i in range(2)]
    #             + [RRTSolver(3000, 1) for i in range(2)],
    #             [PRMSolver(3000, 5) for i in range(3)]
    #             + [RRTSolver(3000, 1) for i in range(3)],
    #         ],
    #         "local_solver": [RRTSolver(60, 0.4)],
    #         "neighborhood_distance": [4],
    #     },
    #     scenes[2],
    #     filename=lambda i, solver: f"./results/HGraph-{i}.png",
    # )

    # plot_results(
    #     "./results/PRMSolver.csv",
    #     "length",
    #     "landmarks",
    #     {"k": 10},
    #     filename="plots/prm_length.png",
    # )
    # plot_results(
    #     "./results/PRMSolver.csv",
    #     "time",
    #     "landmarks",
    #     {"k": 10},
    #     filename="plots/prm_time.png",
    # )
    # plot_results(
    #     "./results/RRTSolver.csv",
    #     "length",
    #     "max_iterations",
    #     {"max_distance": 0.5},
    #     filename="plots/rrt_length.png",
    # )
    # plot_results(
    #     "./results/RRTStarSolver.csv",
    #     "length",
    #     "max_iterations",
    #     {"max_distance": 0.5, "radius": 4},
    #     filename="plots/rrtstar_length.png",
    # )
    # plot_results(
    #     "./results/DRRTSolver.csv",
    #     "length",
    #     "prm_landmarks",
    #     {"max_distance": 0.5, "prm_k": 0, "max_iterations": 2000},
    #     filename="plots/drrt_length.png",
    # )

    # plot_results_comparison(
    #     [os.path.join("results", f) for f in os.listdir("results")],
    #     filename="plots/comparison.png",
    # )
    # plot_results_comparison(
    #     [
    #         os.path.join("results-coffee", f)
    #         for f in os.listdir("results-coffee")
    #     ],
    #     filename="plots/comparison-coffee.png",
    # )



from mrtommy.geometry import Point2D


def distance(positions1: list[Point2D], positions2: list[Point2D]):
    return sum([p1.distance(p2) for p1, p2 in zip(positions1, positions2)])


def nearest_neighbors(
    x: list[Point2D], locations: list[list[Point2D]], k: int
):
    neighbors = sorted(
        locations,
        key=lambda location: distance(x, location),
    )
    return neighbors[:k]



import random

from mrtommy.scene import Scene
from mrtommy.solvers.prm import PRMSolver
from mrtommy.solvers.rrt_star import RRTStarSolver


class DRRTStarSolver(RRTStarSolver):
    def __init__(
        self,
        prm_landmarks: int,
        prm_k: int,
        max_iterations: int,
        max_distance: float,
        radius: float,
    ):
        self.prm_landmarks = prm_landmarks
        self.prm_k = prm_k
        super().__init__(max_iterations, max_distance, radius)

    def get_position(self, scene):
        return [random.choice(roadmap)[0] for roadmap in self.robot_roadmaps]

    def solve(self, scene):
        self.robot_roadmaps = []
        for robot in scene.robots:
            prm = PRMSolver(self.prm_landmarks, self.prm_k)
            prm.solve(Scene([robot], scene.obstacles))
            # TODO: currently the edges generated by PRM are unused
            self.robot_roadmaps.append(list(prm.roadmap.nodes))

        return super().solve(scene)

    def __repr__(self):
        return f"dRRT* with iterations={self.max_iterations}, η={self.max_distance}, radius={self.radius}, landmarks={self.prm_landmarks} and k={self.prm_k}"



import networkx as nx

from mrtommy.geometry import Point2D
from mrtommy.scene import Scene
from mrtommy.solvers.rrt import RRTSolver
from mrtommy.utilities import distance


class RRTStarSolver(RRTSolver):
    def __init__(
        self, max_iterations: int, max_distance: float, radius: float
    ):
        self.radius = radius
        super().__init__(max_iterations, max_distance)

    def solve(self, scene: Scene) -> list[list[Point2D]]:
        self.roadmap = nx.Graph()
        start = tuple(scene.start)
        self.roadmap.add_node(start, cost=0, parent=start)

        for _ in range(self.max_iterations):
            positions = self.get_position(scene)

            closest = min(
                self.roadmap.nodes,
                key=lambda n: distance(positions, n),
            )
            positions = [
                self.move(c, p, self.max_distance)
                for c, p in zip(closest, positions)
            ]

            if not scene.check_edge(closest, positions):
                continue

            # make positions hashable
            positions = tuple(positions)
            cost = self.roadmap.nodes[closest]["cost"] + distance(
                closest, positions
            )

            self.roadmap.add_node(positions, cost=cost, parent=closest)
            self.roadmap.add_edge(
                closest, positions, weight=distance(closest, positions)
            )

            neighbors = nx.ego_graph(
                self.roadmap, positions, self.radius, distance="weight"
            )
            for neighbor in neighbors.nodes:
                neighbor_distance = distance(neighbor, positions)
                neighbor_cost = (
                    self.roadmap.nodes[neighbor]["cost"] + neighbor_distance
                )
                if neighbor_cost < cost:
                    if not scene.check_edge(neighbor, positions):
                        continue
                    cost = neighbor_cost
                    self.roadmap.remove_edge(closest, positions)
                    self.roadmap.add_edge(
                        neighbor, positions, weight=neighbor_distance
                    )
                    self.roadmap.nodes[positions]["parent"] = neighbor
                    closest = neighbor

            for neighbor in neighbors.nodes:
                neighbor_distance = distance(positions, neighbor)
                new_cost = cost + neighbor_distance
                if new_cost < self.roadmap.nodes[neighbor]["cost"]:
                    if not scene.check_edge(neighbor, positions):
                        continue
                    self.roadmap.nodes[neighbor]["cost"] = new_cost
                    self.roadmap.nodes[neighbor]["parent"] = positions
                    self.roadmap.add_edge(
                        positions, neighbor, weight=neighbor_distance
                    )
                    self.roadmap.remove_edge(
                        self.roadmap.nodes[neighbor]["parent"], neighbor
                    )

        end = tuple(scene.end)
        self.roadmap.add_node(end)
        for positions in self.roadmap.nodes:
            if scene.check_edge(positions, end):
                self.roadmap.add_edge(
                    positions, end, weight=distance(positions, end)
                )

        if not nx.algorithms.has_path(self.roadmap, start, end):
            return None

        return nx.algorithms.shortest_path(
            self.roadmap, start, end, weight="weight"
        )

    def __repr__(self):
        return f"RRT* with iterations={self.max_iterations}, η={self.max_distance} and radius={self.radius}"



import itertools as it
import os
import time
from typing import Callable, List, Optional, Type

import matplotlib.pyplot as plt
import pandas as pd
from tqdm import tqdm

from mrtommy.geometry import Point2D
from mrtommy.scene import Scene
from mrtommy.solver import Solver
from mrtommy.utilities import distance


def total_length(path: Optional[list[list[Point2D]]]):
    if path is None:
        return 0

    length = 0
    for i in range(len(path) - 1):
        length += distance(path[i], path[i + 1])
    return length


def run_solver(
    solver_class: Type[Solver],
    options: dict[str, list],
    scene: Scene,
    filename: Optional[
        Callable[
            [
                int,
                Solver,
            ],
            str,
        ]
    ] = None,
):
    items = options.items()
    keys = [item[0] for item in items]
    values = [item[1] for item in items]

    results = pd.DataFrame(columns=["time", "success", "length"] + keys)

    for i, option_values in tqdm(list(enumerate(it.product(*values)))):
        solver_options = {keys[i]: option_values[i] for i in range(len(keys))}
        solver = solver_class(**solver_options)

        start = time.time()
        solution = solver.solve(scene)
        execution_time = time.time() - start

        success = solver.check(scene, solution)
        length = total_length(solution)
        results.loc[i] = [execution_time, success, length] + list(
            map(str, option_values)
        )

        if not filename:
            f = f"./results/{solver_class.__name__}-{'-'.join(map(str, option_values))}.png"
        else:
            f = filename(i, solver)

        solver.plot(
            scene,
            paths=solution,
            filename=f,
        )

    results.to_csv(f"./results/{solver_class.__name__}.csv")


def plot_results(
    result_filename: str,
    y_variable: str,
    x_variable: str,
    other_variable_values: dict[str, str],
    filename: Optional[str] = None,
    display=False,
):
    df = pd.read_csv(result_filename)
    for key in other_variable_values:
        df = df.loc[(df[key] == other_variable_values[key])]

    plt.clf()
    plt.plot(df[x_variable].to_numpy(), df[y_variable].to_numpy())
    plt.title(result_filename.split("/")[-1].split(".")[0])
    plt.xlabel(x_variable)
    plt.ylabel(y_variable)

    if filename is not None:
        plt.savefig(filename)
    if display:
        plt.show()


def plot_results_comparison(
    result_filenames: List[str],
    filename: Optional[str] = None,
    display=False,
):
    plt.clf()
    for f in result_filenames:
        if not f.endswith(".csv"):
            continue

        df = pd.read_csv(f)
        df = df.loc[df["success"]]
        df = df.loc[df["length"] < 100]
        df = df.loc[df["time"] < 100]
        xs = df["time"].to_numpy()
        ys = df["length"].to_numpy()
        plt.scatter(xs, ys, label=os.path.basename(f).split(".")[0])
    plt.legend(loc="upper right")

    if filename is not None:
        plt.savefig(filename)
    if display:
        plt.show()



import itertools as it
import json
import math
import random
from typing import List

from mrtommy.geometry import Point2D, Segment2D


class ObstacleDisc:
    """A disc which robots can't intersect with."""

    location: Point2D
    radius: float

    def __init__(self, location: Point2D, radius: float):
        self.location = location
        self.radius = radius


class RobotDisc:
    """A disc robot which must go from start to end."""

    start: Point2D
    end: Point2D
    radius: float

    def __init__(self, start: Point2D, end: Point2D, radius: float):
        self.start = start
        self.end = end
        self.radius = radius


class Scene:
    """
    Main class to hold Scene information.
    A motion planning scene is a set of robots which need to get from start to end in a
    path of straight lines without intersecting any of the obstacles.
    """

    robots: List[RobotDisc]
    obstacles: List[ObstacleDisc]
    x_min: float
    x_max: float
    y_min: float
    y_max: float

    def __init__(self, robots: list[RobotDisc], obstacles: list[ObstacleDisc]):
        """Initialize a scene from a JSON file containing its information. See the scene JSON schema to learn more about this format."""

        self.robots = robots
        self.obstacles = obstacles

        x_min, y_min = math.inf, math.inf
        x_max, y_max = -math.inf, -math.inf

        for robot in self.robots:
            x_min = min(
                x_min, robot.start.x - robot.radius, robot.end.x - robot.radius
            )
            y_min = min(
                y_min, robot.start.y - robot.radius, robot.end.y - robot.radius
            )
            x_max = max(
                x_max, robot.start.x + robot.radius, robot.end.x + robot.radius
            )
            y_max = max(
                y_max, robot.start.y + robot.radius, robot.end.y + robot.radius
            )

        for obstacle in self.obstacles:
            x_min = min(x_min, obstacle.location.x - obstacle.radius)
            y_min = min(y_min, obstacle.location.y - obstacle.radius)
            x_max = max(x_max, obstacle.location.x + obstacle.radius)
            y_max = max(y_max, obstacle.location.y + obstacle.radius)

        self.x_min = x_min
        self.y_min = y_min
        self.x_max = x_max
        self.y_max = y_max

    def get_boundary(self) -> tuple[float, float, float, float]:
        """Returns x_min, x_max, y_min, y_max of the scene"""

        return self.x_min, self.x_max, self.y_min, self.y_max

    def get_random_positions(self) -> list[Point2D]:
        """Returns a random position sampled at uniform inside the boundary of the scene for each robot"""

        return [
            Point2D(
                random.uniform(self.x_min, self.x_max),
                random.uniform(self.y_min, self.y_max),
            )
            for _ in self.robots
        ]

    def check(self, positions: list[Point2D]):
        """Checks, if the robots are at the given locations (robots[i] at positions[i]), if there are any invalid intersections or not."""

        for robot, position in zip(self.robots, positions):
            for obstacle in self.obstacles:
                if (
                    position.distance(obstacle.location)
                    < obstacle.radius + robot.radius
                ):
                    return False

        for (robot1, position1), (robot2, position2) in it.combinations(
            zip(self.robots, positions), 2
        ):
            if position1.distance(position2) < robot1.radius + robot2.radius:
                return False

        return True

    def get_random_valid_positions(self) -> list[Point2D]:
        p = self.get_random_positions()
        while not self.check(p):
            p = self.get_random_positions()
        return p

    def check_edge(self, start: list[Point2D], end: list[Point2D]):
        """Checks, if the robots are at the given locations (robots[i] at positions[i]), and move in straight lines to the end locations, if there are any invalid intersections or not."""

        assert len(start) == len(end) == len(self.robots)

        for robot_index in range(len(self.robots)):
            for obstacle in self.obstacles:
                if (
                    Segment2D(start[robot_index], end[robot_index]).distance(
                        obstacle.location
                    )
                    < obstacle.radius + self.robots[robot_index].radius
                ):
                    return False

        for r1, r2 in it.combinations(range(len(self.robots)), 2):
            robot1 = self.robots[r1]
            robot2 = self.robots[r2]
            if (
                Segment2D(start[r1], end[r1]).distance(
                    Segment2D(start[r2], end[r2])
                )
                < robot1.radius + robot2.radius
            ):
                return False

        return True

    @property
    def start(self):
        return [r.start for r in self.robots]

    @property
    def end(self):
        return [r.end for r in self.robots]


def load_scene(filename: str):
    with open(filename, "r") as f:
        serialized_scene = json.load(f)

    robots = [
        RobotDisc(Point2D(*r["start"]), Point2D(*r["end"]), r["radius"])
        for r in serialized_scene["robots"]
    ]
    obstacles = [
        ObstacleDisc(Point2D(*o["location"]), o["radius"])
        for o in serialized_scene["obstacles"]
    ]

    return Scene(robots, obstacles)
